@@ -1,4 +1,4 @@
-/* $Id: deliver-append.c,v 1.7 2006/11/23 17:45:32 nicm Exp $ */
+/* $Id: deliver-append.c,v 1.14 2007/03/28 19:59:57 nicm Exp $ */
 
 /*
  * Copyright (c) 2006 Nicholas Marriott <nicm@users.sourceforge.net>
@@ -25,23 +25,28 @@
 #include <unistd.h>
 
 #include "fdm.h"
+#include "deliver.h"
 
-int	 append_deliver(struct deliver_ctx *, struct action *);
-char	*append_desc(struct action *);
+int	 deliver_append_deliver(struct deliver_ctx *, struct actitem *);
+void	 deliver_append_desc(struct actitem *, char *, size_t);
 
-struct deliver deliver_append = { DELIVER_ASUSER, append_deliver, append_desc };
+struct deliver deliver_append = {
+	"append",
+	DELIVER_ASUSER,
+	deliver_append_deliver,
+	deliver_append_desc
+};
 
 int
-append_deliver(struct deliver_ctx *dctx, struct action *t)
+deliver_append_deliver(struct deliver_ctx *dctx, struct actitem *ti)
 {
-	return (do_write(dctx, t, 1));
+	return (do_write(dctx, ti, 1));
 }
 
-char *
-append_desc(struct action *t)
+void
+deliver_append_desc(struct actitem *ti, char *buf, size_t len)
 {
-	char	*s;
+	struct deliver_write_data	*data = ti->data;
 
-	xasprintf(&s, "append \"%s\"", (char *) t->data);
-	return (s);
+	xsnprintf(buf, len, "append \"%s\"", data->path.str);
 }
