@@ -1,4 +1,4 @@
-/* $Id: deliver-smtp.c,v 1.33 2007/01/19 16:53:16 nicm Exp $ */
+/* $Id: deliver-smtp.c,v 1.49 2007/03/28 19:59:57 nicm Exp $ */
 
 /*
  * Copyright (c) 2006 Nicholas Marriott <nicm@users.sourceforge.net>
@@ -26,16 +26,22 @@
 #include <unistd.h>
 
 #include "fdm.h"
+#include "deliver.h"
 
-int	 smtp_deliver(struct deliver_ctx *, struct action *);
-char	*smtp_desc(struct action *);
+int	 deliver_smtp_deliver(struct deliver_ctx *, struct actitem *);
+void	 deliver_smtp_desc(struct actitem *, char *, size_t);
 
-int	 smtp_code(char *);
+int	 deliver_smtp_code(char *);
 
-struct deliver deliver_smtp = { DELIVER_ASUSER, smtp_deliver, smtp_desc };
+struct deliver deliver_smtp = {
+	"smtp",
+	DELIVER_ASUSER,
+	deliver_smtp_deliver,
+	deliver_smtp_desc
+};
 
 int
-smtp_code(char *line)
+deliver_smtp_code(char *line)
 {
 	char		 ch;
 	const char	*errstr;
@@ -57,18 +63,19 @@ smtp_code(char *line)
 }
 
 int
-smtp_deliver(struct deliver_ctx *dctx, struct action *t)
+deliver_smtp_deliver(struct deliver_ctx *dctx, struct actitem *ti)
 {
-	struct account		*a = dctx->account;
-	struct mail		*m = dctx->mail;
-	struct smtp_data	*data = t->data;
-	int		 	 done, code;
-	struct io		*io;
-	char			*cause, *to, *from, *line, *ptr, *lbuf;
-	enum smtp_state		 state;
-	size_t		 	 len, llen;
+	struct account			*a = dctx->account;
+	struct mail			*m = dctx->mail;
+	struct deliver_smtp_data	*data = ti->data;
+	int		 		 done, code;
+	struct io			*io;
+	char				*cause, *to, *from, *line, *ptr, *lbuf;
+	enum deliver_smtp_state		 state;
+	size_t		 		 len, llen;
 
-	io = connectproxy(&data->server, conf.proxy, IO_CRLF, &cause);
+	io = connectproxy(&data->server,
+	    conf.proxy, IO_CRLF, conf.timeout, &cause);
 	if (io == NULL) {
 		log_warnx("%s: %s", a->name, cause);
 		xfree(cause);
@@ -78,10 +85,15 @@ smtp_deliver(struct deliver_ctx *dctx, struct action *t)
 		io->dup_fd = STDOUT_FILENO;
 
 	xasprintf(&from, "%s@%s", conf.info.user, conf.info.host);
-	if (data->to == NULL)
-		to = from;
-	else
-		to = data->to;
+	if (data->to.str == NULL)
+		to = xstrdup(from);
+	else {
+		to = replacestr(&data->to, m->tags, m, &m->rml);
+		if (to == NULL || *to == '\0') {
+			log_warnx("%s: empty to", a->name);
+			goto error;
+		}
+	}
 
 	llen = IO_LINESIZE;
 	lbuf = xmalloc(llen);
@@ -97,7 +109,7 @@ smtp_deliver(struct deliver_ctx *dctx, struct action *t)
 		case -1:
 			goto error;
 		}
-		code = smtp_code(line);
+		code = deliver_smtp_code(line);
 
 		switch (state) {
 		case SMTP_CONNECTING:
@@ -163,6 +175,7 @@ smtp_deliver(struct deliver_ctx *dctx, struct action *t)
 
 	xfree(lbuf);
 	xfree(from);
+	xfree(to);
 
 	io_close(io);
 	io_free(io);
@@ -180,7 +193,10 @@ error:
 	io_flush(io, NULL);
 
 	xfree(lbuf);
-	xfree(from);
+	if (from != NULL)
+		xfree(from);
+	if (to != NULL)
+		xfree(to);
 
 	io_close(io);
 	io_free(io);
@@ -188,14 +204,12 @@ error:
 	return (DELIVER_FAILURE);
 }
 
-char *
-smtp_desc(struct action *t)
+void
+deliver_smtp_desc(struct actitem *ti, char *buf, size_t len)
 {
-	struct smtp_data	*data = t->data;
-	char			*s;
+	struct deliver_smtp_data	*data = ti->data;
 
-	xasprintf(&s, "smtp%s server \"%s\" port %s to \"%s\"",
+	xsnprintf(buf, len, "smtp%s server \"%s\" port %s to \"%s\"",
 	    data->server.ssl ? "s" : "", data->server.host, data->server.port,
-	    data->to);
-	return (s);
+	    data->to.str);
 }
