@@ -1,4 +1,4 @@
-/* $Id: fdm.c,v 1.97 2007/01/24 09:59:37 nicm Exp $ */
+/* $Id: fdm.c,v 1.132 2007/03/30 12:48:04 nicm Exp $ */
 
 /*
  * Copyright (c) 2006 Nicholas Marriott <nicm@users.sourceforge.net>
@@ -30,6 +30,7 @@
 #include <limits.h>
 #include <paths.h>
 #include <pwd.h>
+#include <netdb.h>
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -67,6 +68,16 @@ sighandler(int sig)
 	}
 }
 
+double
+get_time(void)
+{
+	struct timeval	 tv;
+
+	if (gettimeofday(&tv, NULL) != 0)
+		fatal("gettimeofday");
+	return (tv.tv_sec + tv.tv_usec / 1000000.0);
+}
+
 int
 load_conf(void)
 {
@@ -76,7 +87,8 @@ load_conf(void)
 
         yyparse();
 
-        fclose(yyin);
+        if (fclose(yyin) != 0)
+		return (1);
 
         return (0);
 }
@@ -111,6 +123,8 @@ fill_info(const char *home)
 		if (gethostname(host, sizeof host) != 0)
 			fatal("gethostname");
 		conf.info.host = xstrdup(host);
+
+		fill_fqdn(host, &conf.info.fqdn, &conf.info.addr);
 	}
 
 	if (home != NULL && *home != '\0')
@@ -135,13 +149,33 @@ fill_info(const char *home)
 	}
 }
 
+void
+fill_fqdn(char *host, char **fqdn, char **addr)
+{
+	char			 ni[NI_MAXHOST];
+	struct addrinfo		*ai;
+
+	*fqdn = *addr = NULL;
+
+	if (getaddrinfo(host, NULL, NULL, &ai) != 0)
+		return;
+
+	if (getnameinfo(ai->ai_addr,
+	    ai->ai_addrlen, ni, sizeof ni, NULL, 0, NI_NUMERICHOST) == 0)
+		xasprintf(addr, "[%s]", ni);
+
+	if (getnameinfo(ai->ai_addr,
+	    ai->ai_addrlen, ni, sizeof ni, NULL, 0, NI_NAMEREQD) == 0)
+		*fqdn = xstrdup(ni);
+}
+
 int
 dropto(uid_t uid)
 {
 	struct passwd	*pw;
 	gid_t		 gid;
 
-	if (uid == 0)
+	if (uid == NOUSR || uid == 0)
 		return (0);
 
 	pw = getpwuid(uid);
@@ -223,54 +257,106 @@ use_account(struct account *a, char **cause)
 __dead void
 usage(void)
 {
-	printf("usage: %s [-klmnv] [-f conffile] [-u user] [-a name] [-x name] "
-	    "[fetch|poll]\n", __progname);
+	printf("usage: %s [-klmnqv] [-a name] [-D name=value] [-f conffile]"
+	    " [-u user] [-x name] [fetch|poll]\n", __progname);
         exit(1);
 }
 
 int
 main(int argc, char **argv)
 {
-        int		 opt, fds[2], lockfd, status, errors, res;
+        int		 opt, lockfd, status, res;
 	u_int		 i;
 	enum fdmop       op = FDMOP_NONE;
-	const char	*errstr;
-	char		 tmp[512], *ptr, *s;
-	const char	*proxy = NULL;
-	char		*user = NULL, *lock = NULL;
+	const char	*errstr, *proxy = NULL, *s;
+	char		 tmp[BUFSIZ], *ptr, *strs, *user = NULL, *lock = NULL;
 	long		 n;
 	struct utsname	 un;
 	struct passwd	*pw;
 	struct stat	 sb;
-	time_t		 t;
+	time_t		 tt;
 	struct account	*a;
 	pid_t		 pid;
-	struct children	 children;
+	struct children	 children, dead_children;
 	struct child	*child;
 	struct io      **ios, *io;
-	struct timeval	 tv;
 	double		 tim;
 	struct sigaction act;
 	struct msg	 msg;
+	struct msgbuf	 msgbuf;
+	size_t		 off;
+	struct macro	*macro;
+	struct child_fetch_data *cfd;
+#ifdef DEBUG
+	struct rule	*r;
+	struct action	*t;
+#endif
 
 	memset(&conf, 0, sizeof conf);
 	TAILQ_INIT(&conf.accounts);
 	TAILQ_INIT(&conf.rules);
 	TAILQ_INIT(&conf.actions);
 	conf.max_size = DEFMAILSIZE;
+	conf.timeout = DEFTIMEOUT;
 	conf.lock_types = LOCK_FLOCK;
 	conf.impl_act = DECISION_NONE;
 	conf.purge_after = 0;
+	conf.file_umask = DEFUMASK;
+	conf.file_group = NOGRP;
+	conf.queue_high = -1;
+	conf.queue_low = -1;
 
 	log_init(1);
 
 	ARRAY_INIT(&conf.incl);
 	ARRAY_INIT(&conf.excl);
 
-        while ((opt = getopt(argc, argv, "a:f:klmnu:vx:")) != EOF) {
+        while ((opt = getopt(argc, argv, "a:D:f:klmnqu:vx:?")) != EOF) {
                 switch (opt) {
 		case 'a':
-			ARRAY_ADD(&conf.incl, optarg, char *);
+			ARRAY_ADD(&conf.incl, xstrdup(optarg), char *);
+			break;
+		case 'D':
+			if (*optarg != '$' && *optarg != '%') {
+				log_warnx("invalid macro: %s", optarg);
+				exit(1);
+			}
+			ptr = strchr(optarg, '=');
+			if (ptr != NULL)
+				*ptr++ = '\0';
+			if (strlen(optarg) > MAXNAMESIZE) {
+				log_warnx("macro name too long: %s", optarg);
+				exit(1);
+			}
+
+			macro = xmalloc(sizeof *macro);
+			macro->fixed = 1;
+			strlcpy(macro->name, optarg, sizeof macro->name);
+			TAILQ_INSERT_HEAD(&macros, macro, entry);
+
+			if (*optarg == '$') {
+				macro->type = MACRO_STRING;
+				if (ptr == NULL)
+					macro->value.str = xstrdup("");
+				else
+					macro->value.str = xstrdup(ptr);
+				log_debug2("added -D macro \"%s\": %s",
+				    macro->name, macro->value.str);
+				break;
+			}
+
+			macro->type = MACRO_NUMBER;
+			if (ptr == NULL) {
+				macro->value.num = 0;
+				break;
+			}
+			macro->value.num = strtonum(ptr, 0, LLONG_MAX, &errstr);
+			if (errstr != NULL) {
+				log_warnx("number is %s: %s", errstr, ptr);
+				exit(1);
+			}
+			log_debug2("added -D macro \"%s\": %lld", macro->name,
+			    macro->value.num);
 			break;
                 case 'f':
                         conf.conf_file = xstrdup(optarg);
@@ -293,8 +379,11 @@ main(int argc, char **argv)
                 case 'v':
                         conf.debug++;
                         break;
+		case 'q':
+			conf.quiet = 1;
+			break;
 		case 'x':
-			ARRAY_ADD(&conf.excl, optarg, char *);
+			ARRAY_ADD(&conf.excl, xstrdup(optarg), char *);
 			break;
                 case '?':
                 default:
@@ -303,6 +392,8 @@ main(int argc, char **argv)
         }
 	argc -= optind;
 	argv += optind;
+	if (conf.quiet)
+		conf.debug = 0;
 	if (conf.check_only) {
 		if (argc != 0)
 			usage();
@@ -340,22 +431,20 @@ main(int argc, char **argv)
 
 	/* start logging to syslog if necessary */
 	log_init(!conf.syslog);
-	log_debug("version is: %s " BUILD, __progname);
-
-	/* log the start time */
-	t = time(NULL);
-	log_debug("starting at: %.24s", ctime(&t));
+	tt = time(NULL);
+	log_debug("version is: %s " BUILD ", started at: %.24s", __progname,
+	    ctime(&tt));
 
 	/* and the OS version */
 	if (uname(&un) == 0) {
-		log_debug("running on: %s %s %s %s", un.sysname, un.release,
+		log_debug2("running on: %s %s %s %s", un.sysname, un.release,
 		    un.version, un.machine);
 	} else
-		log_debug("running on: (%s)", strerror(errno));
+		log_debug2("uname: %s", strerror(errno));
 
 	/* save the home dir and misc user info */
 	fill_info(getenv("HOME"));
-	log_debug("user is: %s, home is: %s", conf.info.user, conf.info.home);
+	log_debug2("user is: %s, home is: %s", conf.info.user, conf.info.home);
 
 	/* find the config file */
 	if (conf.conf_file == NULL) {
@@ -366,7 +455,7 @@ main(int argc, char **argv)
 			conf.conf_file = xstrdup(SYSCONFFILE);
 		}
 	}
-	log_debug("loading configuration from %s", conf.conf_file);
+	log_debug2("loading configuration from %s", conf.conf_file);
 	if (stat(conf.conf_file, &sb) == -1) {
                 log_warn("%s", conf.conf_file);
 		exit(1);
@@ -377,7 +466,19 @@ main(int argc, char **argv)
                 log_warn("%s", conf.conf_file);
 		exit(1);
 	}
-	log_debug("configuration loaded");
+	log_debug2("configuration loaded");
+
+	/* sort out queue limits */
+	if (conf.queue_high == -1)
+		conf.queue_high = DEFMAILQUEUE;
+	if (conf.queue_low == -1) {
+		conf.queue_low = conf.queue_high * 3 / 4;
+		if (conf.queue_low >= conf.queue_high)
+			conf.queue_low = conf.queue_high - 1;
+ 	}
+
+	/* set the umask */
+	umask(conf.file_umask);
 
 	/* print proxy info */
 	if (conf.proxy != NULL) {
@@ -392,7 +493,7 @@ main(int argc, char **argv)
 			proxy = "SOCKS5";
 			break;
 		}
-		log_debug("using proxy: %s on %s:%s", proxy,
+		log_debug2("using proxy: %s on %s:%s", proxy,
 		    conf.proxy->server.host, conf.proxy->server.port);
 	}
 
@@ -408,7 +509,7 @@ main(int argc, char **argv)
 		if (conf.lock_types & LOCK_DOTLOCK)
 			strlcat(tmp, "dotlock ", sizeof tmp);
 	}
-	log_debug("locking using: %s", tmp);
+	log_debug2("locking using: %s", tmp);
 
 	/* initialise and print headers and domains */
 	if (conf.headers == NULL) {
@@ -417,35 +518,102 @@ main(int argc, char **argv)
 		ARRAY_ADD(conf.headers, xstrdup("to"), char *);
 		ARRAY_ADD(conf.headers, xstrdup("cc"), char *);
 	}
-	s = fmt_strings(NULL, conf.headers);
-	log_debug("headers are: %s", s);
-	xfree(s);
+	strs = fmt_strings(NULL, conf.headers);
+	log_debug2("headers are: %s", strs);
+	xfree(strs);
 	if (conf.domains == NULL) {
 		conf.domains = xmalloc(sizeof *conf.headers);
 		ARRAY_INIT(conf.domains);
-		ARRAY_ADD(conf.domains, conf.info.host, char *);
-	}
-	s = fmt_strings(NULL, conf.domains);
-	log_debug("domains are: %s", s);
-	xfree(s);
-
-	/* save and print tmp dir */
-	conf.tmp_dir = getenv("TMPDIR");
-	if (conf.tmp_dir == NULL || *conf.tmp_dir == '\0')
-		conf.tmp_dir = _PATH_TMP;
-	else {
-		if (stat(conf.tmp_dir, &sb) == -1 || !S_ISDIR(sb.st_mode)) {
-			log_warn("%s", conf.tmp_dir);
-			conf.tmp_dir = _PATH_TMP;
+		ARRAY_ADD(conf.domains, xstrdup(conf.info.host), char *);
+		if (conf.info.fqdn != NULL) {
+			ptr = xstrdup(conf.info.fqdn);
+			ARRAY_ADD(conf.domains, ptr, char *);
+		}
+		if (conf.info.addr != NULL) {
+			ptr = xstrdup(conf.info.addr);
+			ARRAY_ADD(conf.domains, ptr, char *);
 		}
 	}
-	conf.tmp_dir = xstrdup(conf.tmp_dir);
+	strs = fmt_strings(NULL, conf.domains);
+	log_debug2("domains are: %s", strs);
+	xfree(strs);
+
+	/* print the other settings */
+	*tmp = '\0';
+	off = 0;
+	if (conf.allow_many)
+		off = strlcat(tmp, "allow-multiple, ", sizeof tmp);
+	if (conf.no_received)
+		off = strlcat(tmp, "no-received, ", sizeof tmp);
+	if (conf.keep_all)
+		off = strlcat(tmp, "keep-all, ", sizeof tmp);
+	if (conf.del_big)
+		off = strlcat(tmp, "delete-oversized, ", sizeof tmp);
+	if (sizeof tmp > off && conf.purge_after > 0) {
+		off += xsnprintf(tmp + off, (sizeof tmp) - off,
+		    "purge-after=%u, ", conf.purge_after);
+	}
+	if (sizeof tmp > off) {
+		off += xsnprintf(tmp + off, (sizeof tmp) - off,
+		    "maximum-size=%zu, ", conf.max_size);
+	}
+	if (sizeof tmp > off) {
+		off += xsnprintf(tmp + off, (sizeof tmp) - off,
+		    "timeout=%d, ", conf.timeout / 1000);
+	}
+	if (sizeof tmp > off) {
+		off += xsnprintf(tmp + off, (sizeof tmp) - off,
+		    "default-user=%lu, ", (u_long) conf.def_user);
+	}
+	if (sizeof tmp > off && conf.impl_act != DECISION_NONE) {
+		if (conf.impl_act == DECISION_DROP)
+			s = "drop";
+		else if (conf.impl_act == DECISION_KEEP)
+			s = "keep";
+		else
+			s = "none";
+		off += xsnprintf(tmp + off, (sizeof tmp) - off,
+		    "unmatched-mail=%s, ", s);
+	}
+	if (sizeof tmp > off) {
+		off += xsnprintf(tmp + off, (sizeof tmp) - off,
+		    "file-umask=%o%o%o, ", MODE(conf.file_umask));
+	}
+	if (sizeof tmp > off && conf.file_group != NOGRP) {
+		off += xsnprintf(tmp + off, (sizeof tmp) - off,
+		    "file-group=%lu, ", (u_long) conf.file_group);
+	}
+	if (sizeof tmp > off) {
+		off += xsnprintf(tmp + off, (sizeof tmp) - off,
+		    "queue-high=%u, queue-low=%u, ", conf.queue_high,
+		    conf.queue_low);
+	}
+	if (sizeof tmp > off && conf.lock_file != NULL) {
+		off += xsnprintf(tmp + off, (sizeof tmp) - off,
+		    "lock-file=\"%s\", ", conf.lock_file);
+	}
+	if (off >= 2) {
+		tmp[off - 2] = '\0';
+		log_debug2("options are: %s", tmp);
+	}
+
+	/* save and print tmp dir */
+	s = getenv("TMPDIR");
+	if (s == NULL || *s == '\0')
+		s = _PATH_TMP;
+	else {
+		if (stat(s, &sb) == -1 || !S_ISDIR(sb.st_mode)) {
+			log_warn("%s", s);
+			s = _PATH_TMP;
+		}
+	}
+	conf.tmp_dir = xstrdup(s);
 	while ((ptr = strrchr(conf.tmp_dir, '/')) != NULL) {
 		if (ptr == conf.tmp_dir || ptr[1] != '\0')
 			break;
 		*ptr = '\0';
 	}
-	log_debug("using tmp directory: %s", conf.tmp_dir);
+	log_debug2("using tmp directory: %s", conf.tmp_dir);
 
 	/* if -n, bail now, otherwise check there is something to work with */
 	if (conf.check_only)
@@ -520,87 +688,61 @@ main(int argc, char **argv)
         SSL_load_error_strings();
 
 #ifdef DEBUG
-	xmalloc_clear();
 	COUNTFDS("parent");
 #endif
 
 	/* start the children and build the array */
 	ARRAY_INIT(&children);
+	ARRAY_INIT(&dead_children);
+
 	child = NULL;
 	TAILQ_FOREACH(a, &conf.accounts, entry) {
 		if (!use_account(a, NULL))
 			continue;
 
-		if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, fds) != 0)
-			fatal("socketpair");
-		if ((pid = child_fork()) == 0) {
-			for (i = 0; i < ARRAY_LENGTH(&children); i++) {
-				child = ARRAY_ITEM(&children, i,struct child *);
-				io_close(child->io);
-				io_free(child->io);
-				xfree(child);
-			}
-			close(fds[0]);
+		cfd = xmalloc(sizeof *cfd);
+		cfd->account = a;
+		cfd->op = op;
+		cfd->children = &children;
+		child = child_start(&children, conf.child_uid, child_fetch,
+		    parent_fetch, cfd);
 
-			res = do_child(fds[1], op, a);
-#ifdef PROFILE
-			/*
-			 * We want to use _exit rather than exit in the child
-			 * process, and it doesn't run atexit handlers, so run
-			 * _mcleanup manually to force the profiler to dump its
-			 * statistics.
-			 *
-			 * This is icky but it works, and is only for profiling.
-			 */
-			extern void _mcleanup(void);
-			_mcleanup();
-#endif
-			child_exit(res);
-		}
-
-		log_debug("parent: child %ld (%s) started", (long) pid,
+		log_debug2("parent: child %ld (%s) started", (long) child->pid,
 		    a->name);
-
-		child = xcalloc(1, sizeof *child);
-		ARRAY_ADD(&children, child, struct child *);
-		close(fds[1]);
-		child->io = io_create(fds[0], NULL, IO_CRLF);
-		child->pid = pid;
-		child->account = a;
 	}
 
 	if (ARRAY_EMPTY(&children)) {
                 log_warnx("no accounts found");
-		errors = 1;
+		res = 1;
 		goto out;
 	}
 
 #ifndef NO_SETPROCTITLE
 	setproctitle("parent");
 #endif
-	log_debug("parent: started, pid is %ld", (long) getpid());
+	log_debug2("parent: started, pid is %ld", (long) getpid());
+	tim = get_time();
 
-	gettimeofday(&tv, NULL);
-	tim = tv.tv_sec + tv.tv_usec / 1000000.0;
-
-	errors = 0;
-	ios = xcalloc(ARRAY_LENGTH(&children), sizeof (struct io *));
+	res = 0;
+	ios = NULL;
 	while (!ARRAY_EMPTY(&children)) {
 		if (sigint || sigterm)
 			break;
 
 		/* fill the io list */
+		ios = xrealloc(ios, ARRAY_LENGTH(&children), sizeof **ios);
 		for (i = 0; i < ARRAY_LENGTH(&children); i++) {
 			child = ARRAY_ITEM(&children, i, struct child *);
 			ios[i] = child->io;
 		}
 
 		/* poll the io list */
-		switch (io_polln(ios, ARRAY_LENGTH(&children), &io, NULL)) {
+		n = io_polln(ios, ARRAY_LENGTH(&children), &io, INFTIM, NULL);
+		switch (n) {
 		case -1:
-			fatal("parent: child socket error");
+			fatalx("parent: child socket error");
 		case 0:
-			fatalx("child socket closed");
+			fatalx("parent: child socket closed");
 		}
 
 		while (!ARRAY_EMPTY(&children)) {
@@ -614,45 +756,57 @@ main(int argc, char **argv)
 				break;
 
 			/* and handle them if necessary */
-			if (do_parent(child) == 0)
+			if (privsep_recv(child->io, &msg, &msgbuf) != 0)
+				fatalx("parent: privsep_recv error");
+			log_debug3("parent: got message type %d, id %u from "
+			    "child %ld", msg.type, msg.id, (long) child->pid);
+
+			if (child->msg(child, &msg, &msgbuf) == 0)
 				continue;
 
 			/* child has said it is ready to exit, tell it to */
 			memset(&msg, 0, sizeof msg);
 			msg.type = MSG_EXIT;
-			if (privsep_send(child->io, &msg, NULL, 0) != 0)
-				fatalx("child: privsep_send error");
+			if (privsep_send(child->io, &msg, NULL) != 0)
+				fatalx("parent: privsep_send error");
 
-			/* wait for the child  */
+			/* wait for the child */
 			if (waitpid(child->pid, &status, 0) == -1)
 				fatal("waitpid");
 			if (WIFSIGNALED(status)) {
-				errors++;
-				log_debug("parent: child %ld (%s) done, got"
-				    "signal %d", (long) child->pid,
-				    child->account->name, WTERMSIG(status));
+				res = 1;
+				log_debug2("parent: child %ld got signal %d",
+				    (long) child->pid, WTERMSIG(status));
 			} else if (!WIFEXITED(status)) {
-				errors++;
-				log_debug("parent: child %ld (%s) done, didn't"
-				    "exit normally", (long) child->pid,
-				    child->account->name);
+				res = 1;
+				log_debug2("parent: child %ld didn't exit"
+				    "normally", (long) child->pid);
 			} else {
 				if (WEXITSTATUS(status) != 0)
-					errors++;
-				log_debug("parent: child %ld (%s) done, "
-				    "returned %d", (long) child->pid,
-				    child->account->name, WEXITSTATUS(status));
+					res = 1;
+				log_debug2("parent: child %ld returned %d",
+				    (long) child->pid, WEXITSTATUS(status));
 			}
 
-			ARRAY_REMOVE(&children, i, struct child *);
-
 			io_close(child->io);
 			io_free(child->io);
-			xfree(child);
+			child->io = NULL;
+
+			ARRAY_REMOVE(&children, i, struct child *);
+			ARRAY_ADD(&dead_children, child, struct child *);
 		}
 	}
 	xfree(ios);
 
+	/* free the dead children */
+	for (i = 0; i < ARRAY_LENGTH(&dead_children); i++) {
+		child = ARRAY_ITEM(&dead_children, i, struct child *);
+		if (child->data != NULL)
+			xfree(child->data);
+		xfree(child);
+	}
+	ARRAY_FREE(&dead_children);
+
 	if (sigint || sigterm) {
 		act.sa_handler = SIG_IGN;
 		if (sigaction(SIGINT, &act, NULL) < 0)
@@ -686,21 +840,57 @@ main(int argc, char **argv)
 			log_debug2("parent: child %ld killed", (long) pid);
 		}
 
-		errors = 1;
+		res = 1;
 	}
 
-	if (gettimeofday(&tv, NULL) != 0)
-		fatal("gettimeofday");
-	tim = (tv.tv_sec + tv.tv_usec / 1000000.0) - tim;
-	log_debug("parent: finished, total time %.3f seconds", tim);
+	tim = get_time() - tim;
+ 	log_debug2("parent: finished, total time %.3f seconds", tim);
 
 out:
+	if (!conf.allow_many && *conf.lock_file != '\0') {
+		if (unlink(conf.lock_file) != 0)
+			fatal("unlink");
+	}
+
 #ifdef DEBUG
 	COUNTFDS("parent");
+
+	/* free everything */
+	while (!TAILQ_EMPTY(&conf.accounts)) {
+		a = TAILQ_FIRST(&conf.accounts);
+		TAILQ_REMOVE(&conf.accounts, a, entry);
+		free_account(a);
+	}
+	while (!TAILQ_EMPTY(&conf.rules)) {
+		r = TAILQ_FIRST(&conf.rules);
+		TAILQ_REMOVE(&conf.rules, r, entry);
+		free_rule(r);
+	}
+	while (!TAILQ_EMPTY(&conf.actions)) {
+		t = TAILQ_FIRST(&conf.actions);
+		TAILQ_REMOVE(&conf.actions, t, entry);
+		free_action(t);
+	}
+	xfree(conf.info.home);
+	xfree(conf.info.user);
+	xfree(conf.info.uid);
+	xfree(conf.info.host);
+	if (conf.info.fqdn != NULL)
+		xfree(conf.info.fqdn);
+	if (conf.info.addr != NULL)
+		xfree(conf.info.addr);
+	xfree(conf.conf_file);
+	xfree(conf.lock_file);
+	xfree(conf.tmp_dir);
+	free_strings(conf.domains);
+	ARRAY_FREEALL(conf.domains);
+	free_strings(conf.headers);
+	ARRAY_FREEALL(conf.headers);
+	free_strings(&conf.incl);
+	free_strings(&conf.excl);
+
 	xmalloc_report("parent");
 #endif
 
-	if (*conf.lock_file != '\0' && !conf.allow_many)
-		unlink(conf.lock_file);
-	exit(errors);
+	exit(res);
 }
