@@ -1,4 +1,4 @@
-/* $Id: fdm.h,v 1.188 2007/01/25 16:29:25 nicm Exp $ */
+/* $Id: fdm.h,v 1.254 2007/03/29 19:22:30 nicm Exp $ */
 
 /*
  * Copyright (c) 2006 Nicholas Marriott <nicm@users.sourceforge.net>
@@ -42,10 +42,20 @@
 #define SYSCONFFILE	"/etc/fdm.conf"
 #define LOCKFILE	".fdm.lock"
 #define SYSLOCKFILE	"/var/run/fdm.lock"
-#define MAXMAILSIZE	INT_MAX
-#define DEFMAILSIZE	(1 * 1024 * 1024 * 1024)	/* 1 GB */
-#define LOCKSLEEPTIME	2
+#define MAXQUEUEVALUE	50
+#define DEFMAILQUEUE	2
+#define DEFMAILSIZE	(32 * 1024 * 1024)	/* 32 MB */
+#define MAXMAILSIZE	IO_MAXBUFFERLEN
+#define DEFTIMEOUT	(900 * 1000)
+#define LOCKSLEEPTIME	10000
+#define LOCKRETRIES	1000
 #define MAXNAMESIZE	64
+#define DEFUMASK	(S_IRWXG|S_IRWXO)
+#define FILEMODE	(S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH)
+#define DIRMODE		(S_IRWXU|S_IRWXG|S_IRWXO)
+
+#define NOGRP	 	((gid_t) -1)
+#define NOUSR		((uid_t) -1)
 
 extern char	*__progname;
 
@@ -53,6 +63,13 @@ extern char	*__progname;
 #ifndef UID_MAX
 #define UID_MAX UINT_MAX
 #endif
+#ifndef GID_MAX
+#define GID_MAX UINT_MAX
+#endif
+
+#ifndef INFTIM
+#define INFTIM -1
+#endif
 
 #ifndef __dead
 #define __dead __attribute__ ((__noreturn__))
@@ -61,6 +78,7 @@ extern char	*__progname;
 #define __packed __attribute__ ((__packed__))
 #endif
 
+#ifdef DEBUG
 #define NFDS 64
 #define COUNTFDS(s) do {						\
 	int	fd_i, fd_n;						\
@@ -71,6 +89,13 @@ extern char	*__progname;
 	}								\
 	log_debug2("%s: %d file descriptors in use", s, fd_n);		\
 } while (0)
+#endif
+
+/* Convert a file mode. */
+#define MODE(m) \
+	(m & S_IRUSR ? 4 : 0) + (m & S_IWUSR ? 2 : 0) + (m & S_IXUSR ? 1 : 0), \
+    	(m & S_IRGRP ? 4 : 0) +	(m & S_IWGRP ? 2 : 0) +	(m & S_IXGRP ? 1 : 0), \
+	(m & S_IROTH ? 4 : 0) +	(m & S_IWOTH ? 2 : 0) + (m & S_IXOTH ? 1 : 0)
 
 /* Array macros. */
 #define ARRAY_DECL(n, c)						\
@@ -89,6 +114,14 @@ extern char	*__progname;
 	((c *) (a)->list)[(a)->num] = s;				\
 	(a)->num++;							\
 } while (0)
+#define ARRAY_SET(a, i, s, c) do {					\
+	if (((u_int) (i)) >= (a)->num) {				\
+		log_warnx("ARRAY_SET: bad index: %u, at %s:%d",		\
+		    i, __FILE__, __LINE__);				\
+		exit(1);						\
+	}								\
+	((c *) (a)->list)[i] = s;					\
+} while (0)
 #define ARRAY_REMOVE(a, i, c) do {					\
 	if (((u_int) (i)) >= (a)->num) {				\
 		log_warnx("ARRAY_REMOVE: bad index: %u, at %s:%d",	\
@@ -96,13 +129,17 @@ extern char	*__progname;
 		exit(1);						\
 	}								\
 	if (i < (a)->num - 1) {						\
-		c 	*aptr = (a)->list + i;				\
+		c 	*aptr = ((c *) (a)->list) + i;			\
 		memmove(aptr, aptr + 1, (sizeof (c)) * ((a)->num - (i) - 1)); \
 	}								\
 	(a)->num--;							\
         if ((a)->num == 0)						\
 		ARRAY_FREE(a);						\
 } while (0)
+#define ARRAY_EXPAND(a, n, c) do {					\
+	ENSURE_SIZE2((a)->list, (a)->space, (a)->num + n, sizeof (c));	\
+	(a)->num += n;							\
+} while (0)
 #define ARRAY_TRUNC(a, n, c) do {					\
 	if ((a)->num > n)						\
 		(a)->num -= n;				       		\
@@ -148,6 +185,12 @@ extern char	*__progname;
 	(buf) = ensure_for(buf, &(len), size, adj);			\
 } while (0)
 
+/* Description buffer size. */
+#define DESCBUFSIZE 512
+
+/* Replace buffer size. */
+#define REPLBUFSIZE 64
+
 /* Lengths of time. */
 #define TIME_MINUTE 60LL
 #define TIME_HOUR 3600LL
@@ -172,20 +215,18 @@ extern char	*__progname;
 /* Account name match. */
 #define name_match(p, n) (fnmatch(p, n, 0) == 0)
 
-/* Tag match. */
-#define tag_match(p, t) (fnmatch(p, t, 0) == 0)
-
 /* Macros in configuration file. */
 struct macro {
 	char			 name[MAXNAMESIZE];
 	union {
-		long long	 number;
-		char		*string;
+		long long	 num;
+		char		*str;
 	} value;
 	enum {
 		MACRO_NUMBER,
 		MACRO_STRING
 	} type;
+	int			 fixed;
 
 	TAILQ_ENTRY(macro)	entry;
 };
@@ -205,6 +246,21 @@ enum fdmop {
 	FDMOP_FETCH
 };
 
+/*
+ * Wrapper struct for a string that needs tag replacement before it is used.
+ * This is used for anything that needs to be replaced after account and mail
+ * data are available, everything else is replaced at parse time.
+ */
+struct replstr {
+	char		*str;
+};
+ARRAY_DECL(replstrs, struct replstr);
+
+/* Similar to replstr but needs expand_path too. */
+struct replpath {
+	char		*str;
+};
+
 /* Server description. */
 struct server {
 	char		*host;
@@ -230,8 +286,18 @@ struct proxy {
 
 /* Shared memory. */
 struct shm {
+#ifdef SHM_SYSV
+	key_t	 key;
+	int	 id;
+#define SHM_REGISTER(shm)
+#define SHM_DEREGISTER(shm)
+#endif
+#ifdef SHM_MMAP
 	char	 name[MAXPATHLEN];
 	int	 fd;
+#define SHM_REGISTER(shm) cleanup_register((shm)->name)
+#define SHM_DEREGISTER(shm) cleanup_deregister((shm)->name)
+#endif
 
 	void	*data;
 	size_t	 size;
@@ -247,28 +313,122 @@ enum decision {
 	DECISION_KEEP
 };
 
+/* String block entry. */
+struct strbent {
+	size_t	key;
+	size_t	value;
+};
+
+/* String block header. */
+struct strb {
+	u_int		 ent_used;
+	u_int		 ent_max;
+
+	size_t		 str_used;
+	size_t	 	 str_size;
+};
+
+/* Initial string block slots and block size. */
+#define STRBENTRIES 64
+#define STRBBLOCK 1024
+
+/* String block access macros. */
+#define STRB_KEY(sb, sbe) (((char *) (sb)) + (sizeof *(sb)) + (sbe)->key)
+#define STRB_VALUE(sb, sbe) (((char *) (sb)) + (sizeof *(sb)) + (sbe)->value)
+
+#define STRB_ENTRY(sb, n) ((void *) (((char *) (sb)) + \
+	(sizeof *(sb)) + (sb)->str_size + ((n) * (sizeof (struct strbent)))))
+#define STRB_ENTSIZE(sb) ((sb)->ent_max * (sizeof (struct strbent)))
+#define STRB_SIZE(sb) ((sizeof *(sb)) + (sb)->str_size + STRB_ENTSIZE((sb)))
+
+/* Regexp wrapper structs. */
+struct re {
+	char		*str;
+	regex_t		 re;
+	int		 flags;
+};
+
+struct rm {
+	int		 valid;
+
+	size_t		 so;
+	size_t		 eo;
+};
+
+struct rmlist {
+	int		 valid;
+
+	struct rm	 list[NPMATCH];
+};
+
+/* Regexp flags. */
+#define RE_IGNCASE 0x1
+#define RE_NOSUBST 0x2
+
 /* A single mail. */
 struct mail {
-	struct strings		 tags;
-	char			*s;		/* fetch-specific string */
+	u_int			 idx;
+	double			 tim;
+
+	struct strb		*tags;
 
 	struct shm		 shm;
 
 	struct attach		*attach;
+	int			 attach_built;
 
 	char			*base;
 
 	char			*data;
 	size_t			 off;
 
-	size_t	 	 	size;		/* size of mail */
-	size_t	 	 	space;		/* size of malloc'd area */
+	size_t	 	 	 size;		/* size of mail */
+	size_t	 	 	 space;		/* size of malloc'd area */
 
-	ARRAY_DECL(, size_t *)	wrapped;	/* list of wrapped lines */
+	ARRAY_DECL(, size_t *)	 wrapped;	/* list of wrapped lines */
+	char			 wrapchar;	/* wrapped character */
 
 	ssize_t		 	 body;		/* offset of body */
+
+	/* XXX move below into special struct and just cp it in mail_*? */
+	struct rmlist		 rml;		/* regexp matches */
+
+	enum decision		 decision;	/* final deliver decision */
+
+	void			 (*auxfree)(void *);
+	void			*auxdata;
 };
 
+/* Mail fetch/delivery return codes. */
+#define MAIL_CONTINUE 0
+#define MAIL_DELIVER 1
+#define MAIL_MATCH 2
+#define MAIL_ERROR 3
+#define MAIL_BLOCKED 4
+#define MAIL_DONE 5
+
+/* Mail fetch/delivery context. */
+struct mail_ctx {
+	int				 done;
+	u_int				 msgid;
+
+	struct account			*account;
+	struct io			*io;
+	struct mail			*mail;
+
+	struct rule			*rule;
+	ARRAY_DECL(, struct rule *)	 stack;
+	struct expritem			*expritem;
+	int				 result;
+	int				 matched;
+
+	TAILQ_HEAD(, deliver_ctx)	 dqueue;
+
+	TAILQ_ENTRY(mail_ctx)		 entry;
+};
+TAILQ_HEAD(mail_queue, mail_ctx);
+extern struct mail_queue mail_queue;
+
 /* An attachment. */
 struct attach {
 	u_int	 	 	 idx;
@@ -286,27 +446,94 @@ struct attach {
 	TAILQ_ENTRY(attach)	 entry;
 };
 
-/* Regexp wrapper struct. */
-struct re {
-	char		*s;
-	regex_t		 re;
+/* Privsep message types. */
+enum msgtype {
+	MSG_ACTION,
+	MSG_EXIT,
+	MSG_DONE,
+	MSG_COMMAND
+};
+
+/* Privsep message data. */
+struct msgdata {
+	int	 		 	 error;
+	struct mail		 	 mail;
+
+	/* these only work so long as they aren't moved in either process */
+	struct account			*account;
+	struct actitem			*actitem;
+	struct match_command_data	*cmddata;
+
+	uid_t			 	 uid;
+};
+
+/* Privsep message buffer. */
+struct msgbuf {
+	void		*buf;
+	size_t		 len;
+};
+
+/* Privsep message. */
+struct msg {
+	u_int		 id;
+	enum msgtype	 type;
+	size_t		 size;
+
+	struct msgdata	 data;
 };
 
 /* A single child. */
 struct child {
 	pid_t		 pid;
 	struct io	*io;
-	struct account	*account;
+
+	void		*data;
+	int		 (*msg)(struct child *, struct msg *, struct msgbuf *);
+
+	void		*buf;
+	size_t		 len;
 };
 
 /* List of children. */
 ARRAY_DECL(children, struct child *);
 
+/* Fetch child data. */
+struct child_fetch_data {
+	struct account	*account;
+	enum fdmop	 op;
+	struct children	*children;
+};
+
+/* Deliver child data. */
+struct child_deliver_data {
+	void			 (*hook)(int, struct account *, struct msg *,
+				      struct child_deliver_data *, int *);
+
+	struct child 		*child; /* the source of the request */
+
+	u_int			 msgid;
+	const char		*name;
+
+	struct account		*account;
+	struct mail		*mail;
+	struct actitem		*actitem;
+
+	struct deliver_ctx	*dctx;
+	struct mail_ctx		*mctx;
+
+	struct match_command_data *cmddata;
+};
+
+/* Users list. */
+ARRAY_DECL(users, uid_t);
+
 /* Account entry. */
 struct account {
+	u_int			 idx;
+
 	char			 name[MAXNAMESIZE];
 
-	struct strings		*users;
+	struct users		*users;
 	int			 find_uid;
 
 	int			 disabled;
@@ -318,15 +545,27 @@ struct account {
 	TAILQ_ENTRY(account)	 entry;
 };
 
+/* Action item. */
+struct actitem {
+	u_int			 idx;
+
+	struct deliver		*deliver;
+	void			*data;
+
+	TAILQ_ENTRY(actitem)	 entry;
+}; 
+
+/* Action list. */
+TAILQ_HEAD(actlist, actitem);
+
 /* Action definition. */
 struct action {
 	char			 name[MAXNAMESIZE];
 
-	struct strings		*users;
+	struct users		*users;
 	int			 find_uid;
 
-	struct deliver		*deliver;
-	void			*data;
+	struct actlist		*list;
 
 	TAILQ_ENTRY(action)	 entry;
 };
@@ -365,28 +604,21 @@ TAILQ_HEAD(expr, expritem);
 /* Rule list. */
 TAILQ_HEAD(rules, rule);
 
-/* Rule types. */
-enum ruletype {
-	RULE_EXPRESSION,
-	RULE_ALL
-};
-
 /* Rule entry. */
 struct rule {
-	enum ruletype		 type;
+	u_int			 idx;
 
 	struct strings		*accounts;
 	struct expr		*expr;
 
-	struct strings		*users;
+	struct users		*users;
 	int			 find_uid;	/* find uids from headers */
 
 	int			 stop;		/* stop matching at this rule */
 
-	char			*tag;
-
 	struct rules		 rules;
-	struct strings		*actions;
+	struct action		*lambda;
+	struct replstrs		*actions;
 
 	TAILQ_ENTRY(rule)	 entry;
 };
@@ -398,12 +630,13 @@ struct rule {
 
 /* Configuration settings. */
 struct conf {
+	int			 quiet;
 	int 			 debug;
 	int			 syslog;
 
 	uid_t			 child_uid;
 	gid_t			 child_gid;
-	const char		*tmp_dir;
+	char			*tmp_dir;
 
 	struct strings	 	 incl;
 	struct strings		 excl;
@@ -421,6 +654,8 @@ struct conf {
 		char		*user;
 		char		*uid;
 		char		*host;
+		char		*fqdn;
+		char		*addr;
 	} info;
 
 	char			*conf_file;
@@ -428,10 +663,18 @@ struct conf {
 	int			 check_only;
 	int			 allow_many;
 	int			 keep_all;
+	int			 no_received;
 	u_int			 purge_after;
 	enum decision		 impl_act;
 
+	int			 queue_high;
+	int			 queue_low;
+
+	mode_t			 file_umask;
+	gid_t			 file_group;
+
 	size_t			 max_size;
+	int			 timeout;
 	int		         del_big;
 	u_int			 lock_types;
 	uid_t			 def_user;
@@ -464,7 +707,7 @@ extern struct conf		 conf;
 #define IO_FLUSHSIZE (2 * IO_BLOCKSIZE)
 
 /* Maximum number of pollfds. */
-#define IO_POLLFDS 64
+#define IO_POLLFDS 256
 
 /* IO buffer size macros. */
 #define IO_ROUND(n) (((n / IO_BLOCKSIZE) + 1) * IO_BLOCKSIZE)
@@ -486,6 +729,7 @@ struct io {
 #define IO_NEEDFILL 0x4
 #define IO_NEEDPUSH 0x8
 #define IO_FIXED 0x10			/* fixed write buffer */
+#define IO_NOWAIT 0x20			/* don't block on poll */
 
 	char		*rbase;		/* buffer start */
 	size_t		 rspace;	/* total size of buffer */
@@ -500,11 +744,12 @@ struct io {
 	char		*lbuf;		/* line buffer */
 	size_t		 llen;		/* line buffer size */
 
+	int		 timeout;
 	const char	*eol;
 };
 
 /* Command flags. */
-#define CMD_IN  0x1 
+#define CMD_IN  0x1
 #define CMD_OUT 0x2
 #define CMD_ONCE 0x4
 
@@ -513,130 +758,13 @@ struct cmd {
 	pid_t	 	 pid;
 	int		 status;
 	int		 flags;
+	int		 timeout;
 
 	struct io	*io_in;
 	struct io	*io_out;
 	struct io	*io_err;
 };
 
-/* Poll return codes. */
-#define POLL_SUCCESS FETCH_SUCCESS
-#define POLL_ERROR FETCH_ERROR
-
-/* Fetch return codes. */
-#define FETCH_SUCCESS 0
-#define FETCH_ERROR 1
-#define FETCH_OVERSIZE 2
-#define FETCH_COMPLETE 3
-
-/* Fetch functions. */
-struct fetch {
-	const char	*ports[2];	/* normal port, ssl port */
-
-	int		 (*init)(struct account *);
-	int	 	 (*connect)(struct account *);
-	int 		 (*poll)(struct account *, u_int *);
-	int	 	 (*fetch)(struct account *, struct mail *);
-	int		 (*purge)(struct account *);
-	int		 (*delete)(struct account *);
-	int		 (*keep)(struct account *);
-	int		 (*disconnect)(struct account *);
-	int		 (*free)(struct account *);
-	char		*(*desc)(struct account *);
-};
-
-/* Deliver context. */
-struct deliver_ctx {
-	struct account	*account;
-	struct mail	*mail;
-
-	struct mail	 wr_mail;
-
-	enum decision	*decision;
-
-	int		 pmatch_valid;
-	regmatch_t	 pmatch[NPMATCH];
-};
-
-/* Deliver return codes. */
-#define DELIVER_SUCCESS 0
-#define DELIVER_FAILURE 1
-
-/* Delivery types. */
-enum delivertype {
-	DELIVER_INCHILD,	/* do not pass up to parent */
-	DELIVER_ASUSER,		/* do pass up to parent to drop privs */
-	DELIVER_WRBACK		/* modifies mail: pass up to parent and expect
-				   a new mail back */
-};
-
-/* Deliver functions. */
-struct deliver {
-	enum delivertype type;
-
-	int	 	 (*deliver)(struct deliver_ctx *, struct action *);
-	char		*(*desc)(struct action *);
-};
-
-#define MATCH_FALSE 0
-#define MATCH_TRUE 1
-#define MATCH_ERROR 2
-
-/* Match context. */
-struct match_ctx {
-	struct io	*io;
-	struct account	*account;
-	struct mail     *mail;
-
-	enum decision	 decision;
-
-	int		 matched;
-	int		 stopped;
-
-	int		 pmatch_valid;
-	regmatch_t	 pmatch[NPMATCH];
-};
-
-/* Match functions. */
-struct match {
-	int		 (*match)(struct match_ctx *, struct expritem *);
-	char 		*(*desc)(struct expritem *);
-};
-
-/* Privsep message types. */
-enum msgtype {
-	MSG_ACTION,
-	MSG_EXIT,
-	MSG_DONE,
-	MSG_COMMAND
-};
-
-/* Privsep message data. */
-struct msgdata {
-	int	 	 	 error;
-	struct mail	 	 mail;
-
-	int		 	 pmatch_valid;
-	regmatch_t	 	 pmatch[NPMATCH];
-
-	/* these only work so long as they aren't moved in either process */
-	struct account		*account;
-	struct action		*action;
-	struct command_data	*cmddata;
-
-	uid_t		 	 uid;
-};
-
-/* Privsep message. */
-struct msg {
-	u_int		 n;
-
-	enum msgtype	 type;
-	size_t		 size;
-
-	struct msgdata	 data;
-};
-
 /* Comparison operators. */
 enum cmp {
 	CMP_EQ,
@@ -645,267 +773,6 @@ enum cmp {
 	CMP_GT
 };
 
-/* Match attachment data. */
-struct attachment_data {
-	enum {
-		ATTACHOP_COUNT,
-		ATTACHOP_TOTALSIZE,
-		ATTACHOP_ANYSIZE,
-		ATTACHOP_ANYTYPE,
-		ATTACHOP_ANYNAME
-	} op;
-
-	enum cmp	 	 cmp;
-	union {
-		size_t		 size;
-		long long	 number;
-		char		*string;
-		struct re	 re;
-	} value;
-};
-
-/* Match age data. */
-struct age_data {
-	long long	 time;
-	enum cmp	 cmp;
-};
-
-/* Match size data. */
-struct size_data {
-	size_t		 size;
-	enum cmp	 cmp;
-};
-
-/* Match tagged data. */
-struct tagged_data {
-	char		*tag;
-};
-
-/* Match string data. */
-struct string_data {
-	struct re	 re;
-
-	char		*s;
-};
-
-/* Match regexp data. */
-struct regexp_data {
-	struct re	 re;
-
-	enum area 	 area;
-};
-
-/* Match command data. */
-struct command_data {
-	char		*cmd;
-	uid_t		 uid;
-	int		 pipe;		/* pipe mail to command */
-
-	struct re	 re;		/* re->re NULL to not check */
-	int		 ret;		/* -1 to not check */
-};
-
-/* Fetch maildir data. */
-struct maildir_data {
-	struct strings	*maildirs;
-
-	struct strings	*paths;
-	u_int		 index;
-
-	DIR		*dirp;
-	char		*path;
-	char		 entry[MAXPATHLEN];
-};
-
-/* NNTP group entry. */
-struct nntp_group {
-	char		*name;
-	int		 ignore;
-
-	u_int		 size;
-	u_int		 last;
-	char		*id;
-};
-
-/* NNTP group list. */
-ARRAY_DECL(nntp_groups, struct nntp_group *);
-
-/* Fetch nntp data. */
-struct nntp_data {
-	char		*path;
-
-	struct server	 server;
-	struct strings	*names;
-
-	u_int		 group;
-	struct nntp_groups groups;
-
-	struct io	*io;
-};
-#define GET_GROUP(d, i) ARRAY_ITEM(&d->groups, i, struct nntp_group *)
-#define CURRENT_GROUP(d) GET_GROUP(d, d->group)
-#define TOTAL_GROUPS(d) ARRAY_LENGTH(&d->groups)
-#define ADD_GROUP(d, g) ARRAY_ADD(&d->groups, g, struct nntp_group *)
-
-/* Fetch stdin data. */
-struct stdin_data {
-	int		 complete;
-
-	struct io	*io;
-};
-
-/* Fetch pop3 data. */
-struct pop3_data {
-	char		*user;
-	char		*pass;
-
-	struct server	 server;
-
-	u_int		 cur;
-	u_int		 num;
-
-	char		*uid;
-	struct strings	 kept;
-
-	struct io	*io;
-};
-
-/* IMAP tag types. */
-#define IMAP_TAG_NONE -1
-#define IMAP_TAG_CONTINUE -2
-#define IMAP_TAG_ERROR -3
-
-/* IMAP line types. */
-#define IMAP_TAGGED 0
-#define IMAP_CONTINUE 1
-#define IMAP_UNTAGGED 2
-#define IMAP_RAW 3
-
-/* Fetch imap data. */
-struct imap_data {
-	struct server	 server;
-	char		*pipecmd;
-
-	struct io	*io;
-	struct cmd	*cmd;
-
-	char		*user;
-	char		*pass;
-	char		*folder;
-
-	int		 tag;
-	u_int		 cur;
-	u_int		 num;
-
-	u_int	 	 uid;
-	ARRAY_DECL(, u_int) kept;
-
-	char		*s;
-
-	size_t		 llen;
-	char		*lbuf;
-
-	char		*(*getln)(struct account *a, int);
-	int		 (*putln)(struct account *a, const char *, ...);
-	void		 (*flush)(struct account *a);	
-};
-
-/* Deliver smtp states. */
-enum smtp_state {
-	SMTP_CONNECTING,
-	SMTP_HELO,
-	SMTP_FROM,
-	SMTP_TO,
-	SMTP_DATA,
-	SMTP_DONE,
-	SMTP_QUIT
-};
-
-/* Deliver smtp data. */
-struct smtp_data {
-	struct server	 server;
-	char		*to;
-};
-
-/* Deliver mbox data. */
-struct mbox_data {
-	char		*path;
-	int		 compress;
-};
-
-/* match-age.c */
-extern struct match	 match_age;
-
-/* match-attachment.c */
-extern struct match	 match_attachment;
-
-/* match-matched.c */
-extern struct match	 match_matched;
-
-/* match-unmatched.c */
-extern struct match	 match_unmatched;
-
-/* match-size.c */
-extern struct match	 match_size;
-
-/* match-tagged.c */
-extern struct match	 match_tagged;
-
-/* match-string.c */
-extern struct match	 match_string;
-
-/* match-command.c */
-extern struct match	 match_command;
-
-/* match-regexp.c */
-extern struct match	 match_regexp;
-
-/* fetch-maildir.c */
-extern struct fetch 	 fetch_maildir;
-
-/* fetch-stdin.c */
-extern struct fetch 	 fetch_stdin;
-
-/* fetch-nntp.c */
-extern struct fetch 	 fetch_nntp;
-
-/* fetch-pop3.c */
-extern struct fetch 	 fetch_pop3;
-
-/* fetch-imap.c */
-extern struct fetch 	 fetch_imap;
-
-/* fetch-imappipe.c */
-extern struct fetch 	 fetch_imappipe;
-
-/* deliver-smtp.c */
-extern struct deliver	 deliver_smtp;
-
-/* deliver-pipe.c */
-extern struct deliver 	 deliver_pipe;
-
-/* deliver-drop.c */
-extern struct deliver 	 deliver_drop;
-
-/* deliver-keep.c */
-extern struct deliver 	 deliver_keep;
-
-/* deliver-maildir.c */
-extern struct deliver 	 deliver_maildir;
-
-/* deliver-mbox.c */
-extern struct deliver 	 deliver_mbox;
-
-/* deliver-write.c */
-extern struct deliver 	 deliver_write;
-int	 		 do_write(struct deliver_ctx *, struct action *, int);
-
-/* deliver-append.c */
-extern struct deliver 	 deliver_append;
-
-/* deliver-rewrite.c */
-extern struct deliver 	 deliver_rewrite;
-
 #ifdef NO_SETRESUID
 #define setresuid(r, e, s) setreuid(r, e)
 #endif
@@ -932,62 +799,60 @@ size_t	 		 strlcat(char *, const char *, size_t);
 
 #ifdef NO_ASPRINTF
 /* asprintf.c */
-int			 asprintf(char **, const char *, ...);
+int 			 asprintf(char **, const char *, ...);
 int			 vasprintf(char **, const char *, va_list);
 #endif
 
 /* shm.c */
-void 			*shm_reopen(struct shm *);
-void			*shm_malloc(struct shm *, size_t);
-void			*shm_realloc(struct shm *, size_t, size_t);
-void			 shm_free(struct shm *);
+void			*shm_create(struct shm *, size_t);
+int			 shm_owner(struct shm *, uid_t, gid_t);
 void			 shm_destroy(struct shm *);
+void			 shm_close(struct shm *);
+void			*shm_reopen(struct shm *);
+void			*shm_resize(struct shm *, size_t, size_t);
+
+/* lex.l */
+extern char		*curfile;
+void			 include_start(char *);
+int			 include_finish(void);
 
 /* parse.y */
+extern struct strb	*parse_tags;
 extern struct macros	 macros;
+struct users		*weed_users(struct users *);
 struct strings 		*weed_strings(struct strings *);
+void			 free_strings(struct strings *);
 char 			*fmt_strings(const char *, struct strings *);
+char 			*fmt_users(const char *, struct users *);
 struct macro		*find_macro(char *);
 struct actions		*match_actions(char *);
+void			 free_action(struct action *);
+void			 free_rule(struct rule *);
+void			 free_account(struct account *);
+char			*expand_path(char *);
+
+/* netrc.c */
+FILE 			*netrc_open(const char *, char **);
+void			 netrc_close(FILE *);
+int			 netrc_lookup(FILE *, const char *, char **, char **);
 
 /* fdm.c */
+double			 get_time(void);
 int			 dropto(uid_t);
 int			 check_incl(char *);
 int		         check_excl(char *);
 int			 use_account(struct account *, char **);
 void			 fill_info(const char *);
-
-/* cache.c */
-struct cache 		*cache_open(char *, char **);
-void			 cache_close(struct cache *);
-u_int			 cache_compact(struct cache *, long long, u_int *);
-void			 cache_add(struct cache *, char *);
-void			 cache_update(struct cache *, char *);
-void			 cache_put(struct cache *, char *, u_int);
-int			 cache_get(struct cache *, char *, u_int *);
-int			 cache_contains(struct cache *, char *);
-
-/* imap-common.c */
-int			 imap_tag(char *);
-int			 imap_init(struct account *);
-int			 imap_free(struct account *);
-int			 imap_login(struct account *);
-int			 imap_select(struct account *);
-int			 imap_close(struct account *);
-int			 imap_logout(struct account *);
-void			 imap_abort(struct account *);
-int			 imap_uid(struct account *);
-int			 imap_poll(struct account *, u_int *);
-int			 imap_fetch(struct account *, struct mail *);
-int			 imap_purge(struct account *);
-int			 imap_delete(struct account *);
-int	 		 imap_keep(struct account *);
+void			 fill_fqdn(char *, char **, char **);
 
 /* re.c */
 int			 re_compile(struct re *, char *, int, char **);
-int			 re_execute(struct re *, char *, int, regmatch_t *,
-			     int, char **);
+int			 re_string(struct re *, char *, struct rmlist *,
+			     char **);
+int			 re_block(struct re *, void *, size_t, struct rmlist *,
+			     char **);
 int			 re_simple(struct re *, char *, char **);
+void			 re_free(struct re *);
 
 /* attach.c */
 struct attach 		*attach_visit(struct attach *, u_int *);
@@ -996,14 +861,15 @@ struct attach 		*attach_build(struct mail *);
 void			 attach_free(struct attach *);
 
 /* privsep.c */
-int			 privsep_send(struct io *, struct msg *, void *,
-			     size_t);
+int			 privsep_send(struct io *, struct msg *,
+			     struct msgbuf *);
 int			 privsep_check(struct io *);
-int			 privsep_recv(struct io *, struct msg *, void **,
-			     size_t *);
+int			 privsep_recv(struct io *, struct msg *,
+			     struct msgbuf *);
 
 /* command.c */
-struct cmd 		*cmd_start(const char *, int, char *, size_t, char **);
+struct cmd 		*cmd_start(const char *, int, int, char *, size_t,
+			     char **);
 int			 cmd_poll(struct cmd *, char **, char **, char **,
 			     size_t *, char **);
 void			 cmd_free(struct cmd *);
@@ -1011,38 +877,65 @@ void			 cmd_free(struct cmd *);
 /* child.c */
 int			 child_fork(void);
 __dead void		 child_exit(int);
-int			 do_child(int, enum fdmop, struct account *);
+struct child 		*child_start(struct children *, uid_t,
+    			     int (*)(struct child *, struct io *),
+			     int (*)(struct child *, struct msg *,
+    			     struct msgbuf *), void *);
 
-/* parent.c */
-int			 do_parent(struct child *);
+/* child-fetch.c */
+int			 child_fetch(struct child *, struct io *);
+
+/* child-deliver.c */
+int			 child_deliver(struct child *, struct io *);
+void			 child_deliver_action_hook(int, struct account *,
+			     struct msg *, struct child_deliver_data *, int *);
+void			 child_deliver_cmd_hook(int, struct account *,
+			     struct msg *, struct child_deliver_data *, int *);
+
+/* parent-fetch.c */
+int			 parent_fetch(struct child *, struct msg *,
+			     struct msgbuf *);
+
+/* parent-deliver.c */
+int			 parent_deliver(struct child *, struct msg *,
+			     struct msgbuf *);
 
 /* connect.c */
 struct proxy 		*getproxy(const char *);
 struct io 		*connectproxy(struct server *, struct proxy *,
-			     const char *, char **);
-struct io		*connectio(struct server *, const char *, char **);
+			     const char *, int, char **);
+struct io		*connectio(struct server *, const char *, int, char **);
 
 /* mail.c */
-void			 mail_open(struct mail *, size_t);
+int			 mail_open(struct mail *, size_t);
 void			 mail_send(struct mail *, struct msg *);
-void			 mail_receive(struct mail *, struct msg *);
-void			 mail_reopen(struct mail *, char *);
+int			 mail_receive(struct mail *, struct msg *, int);
 void			 mail_close(struct mail *);
 void			 mail_destroy(struct mail *);
-void			 resize_mail(struct mail *, size_t);
-int 			 makepath(char *, size_t, char *, char *);
-int			 openlock(char *, u_int, int, mode_t);
-void			 closelock(int, char *, u_int);
+int			 mail_resize(struct mail *, size_t);
+char 			*rfc822_time(time_t, char *, size_t);
+int printflike3		 printpath(char *, size_t, const char *, ...);
+int			 openlock(const char *, u_int, int, mode_t);
+void			 closelock(int, const char *, u_int);
+int			 checkperms(const char *, const char *, int *);
 void			 line_init(struct mail *, char **, size_t *);
 void			 line_next(struct mail *, char **, size_t *);
-char 			*find_header(struct mail *, const char *, size_t *);
-struct strings		*find_users(struct mail *);
+int printflike3		 insert_header(struct mail *, const char *,
+			     const char *, ...);
+int			 remove_header(struct mail *, const char *);
+char 			*find_header(struct mail *, const char *, size_t *,
+			     int);
+struct users		*find_users(struct mail *);
 char			*find_address(char *, size_t, size_t *);
 void			 trim_from(struct mail *);
 char 		        *make_from(struct mail *);
 u_int			 fill_wrapped(struct mail *);
 void			 set_wrapped(struct mail *, char);
 
+/* mail-state.c */
+int	mail_match(struct mail_ctx *, struct msg *, struct msgbuf *);
+int	mail_deliver(struct mail_ctx *, struct msg *, struct msgbuf *);
+
 /* cleanup.c */
 void			 cleanup_check(void);
 void			 cleanup_flush(void);
@@ -1050,25 +943,40 @@ void			 cleanup_purge(void);
 void			 cleanup_register(char *);
 void			 cleanup_deregister(char *);
 
+/* strb.c */
+void		 	 strb_create(struct strb **);
+void			 strb_clear(struct strb **);
+void			 strb_destroy(struct strb **);
+void			 strb_dump(struct strb *, const char *,
+    			     void (*)(const char *, ...));
+void printflike3	 strb_add(struct strb **, const char *, const char *,
+			     ...);
+void			 strb_vadd(struct strb **, const char *, const char *,
+			     va_list);
+struct strbent 		*strb_find(struct strb *, const char *);
+struct strbent	 	*strb_match(struct strb *, const char *);
+
 /* replace.c */
-#define REPL_LEN 62
-#define REPL_IDX(ch) /* LINTED */ 				\
-	(((ch) >= 'a' && (ch) <= 'z') ? (ch) - 'a' :       	\
-	(((ch) >= 'A' && (ch) <= 'Z') ? 26 + (ch) - 'A' : 	\
-	(((ch) >= '0' && (ch) <= '9') ? 52 + (ch) - '0' : -1)))
-char 			*replacepmatch(char *, struct account *,
-    			     struct action *, char *, struct mail *,
-    			     int, regmatch_t [NPMATCH]);
-char			*replaceinfo(char *, struct account *, struct action *,
-    			     char *);
-char 			*replace(char *, char *[REPL_LEN]);
+void printflike3	 add_tag(struct strb **, const char *, const char *,
+			     ...);
+const char 		*find_tag(struct strb *, const char *);
+const char		*match_tag(struct strb *, const char *);
+void			 default_tags(struct strb **, char *, struct account *);
+void			 update_tags(struct strb **);
+char 			*replace(char *, struct strb *, struct mail *,
+			     struct rmlist *);
+char 			*replacestr(struct replstr *, struct strb *,
+			     struct mail *, struct rmlist *);
+char 			*replacepath(struct replpath *, struct strb *,
+    			     struct mail *, struct rmlist *);
 
 /* io.c */
-struct io		*io_create(int, SSL *, const char *);
+struct io		*io_create(int, SSL *, const char *, int);
 void			 io_free(struct io *);
 void			 io_close(struct io *);
 int			 io_update(struct io *, char **);
-int			 io_polln(struct io **, u_int, struct io **, char **);
+int			 io_polln(struct io **, u_int, struct io **, int,
+			     char **);
 int			 io_poll(struct io *, char **);
 int			 io_read2(struct io *, void *, size_t);
 void 			*io_read(struct io *, size_t);
@@ -1085,8 +993,8 @@ int			 io_flush(struct io *, char **);
 int			 io_wait(struct io *, size_t, char **);
 
 /* log.c */
+void			 vlog(FILE *, int, const char *, va_list);
 void			 log_init(int);
-void		    	 vlog(int, const char *, va_list);
 void printflike1	 log_warn(const char *, ...);
 void printflike1	 log_warnx(const char *, ...);
 void printflike1	 log_info(const char *, ...);
@@ -1100,13 +1008,16 @@ __dead void		 fatalx(const char *);
 void		*ensure_size(void *, size_t *, size_t, size_t);
 void		*ensure_for(void *, size_t *, size_t, size_t);
 char		*xstrdup(const char *);
-int printflike3	 xsnprintf(char *, size_t, const char *, ...);
 void		*xxcalloc(size_t, size_t);
 void		*xxmalloc(size_t);
 void		*xxrealloc(void *, size_t, size_t);
 void		 xxfree(void *);
 int printflike2	 xxasprintf(char **, const char *, ...);
 int		 xxvasprintf(char **, const char *, va_list);
+int printflike3	 xsnprintf(char *, size_t, const char *, ...);
+int		 xvsnprintf(char *, size_t, const char *, va_list);
+char 		*xdirname(const char *);
+char 		*xbasename(const char *);
 
 /* xmalloc-debug.c */
 #ifdef DEBUG
@@ -1130,6 +1041,7 @@ int		 dxvasprintf(const char *, u_int, char **, const char *,
 #define xrealloc(p, n, s) dxrealloc(__FILE__, __LINE__, p, n, s)
 #define xfree(p) dxfree(__FILE__, __LINE__, p)
 #define xasprintf(pp, ...) dxasprintf(__FILE__, __LINE__, pp, __VA_ARGS__)
+#define xvasprintf(pp, fmt, ap) dxvasprintf(__FILE__, __LINE__, pp, fmt, ap)
 #else
 #define xmalloc(s) xxmalloc(s)
 #define xcalloc(n, s) xxcalloc(n, s)
